import numpy
from numpy import linalg

import logging; logger = logging.getLogger("underworlds.helpers.geometry")

from ..types import MESH
from functools import reduce

def transform(vector3, matrix4x4):
    """ Apply a transformation matrix on a 3D vector.

    :param vector3: a numpy array with 3 elements
    :param matrix4x4: a numpy 4x4 matrix
    """
    return numpy.dot(matrix4x4, numpy.append(vector3, 1.))

def get_scene_bounding_box(scene):
    """
    Returns the axis-aligned bounding box (AABB) of a whole scene,
    or None if the scene is empty.
    """
    bb_min = [1e10, 1e10, 1e10] # x,y,z
    bb_max = [-1e10, -1e10, -1e10] # x,y,z

    if not scene.rootnode.children:
        logger.warning("rootnode has no children! The scene is probably empty.")
        return None, None

    return _compute_bounding_box_for_node(scene.nodes, scene.rootnode, bb_min, bb_max, linalg.inv(scene.rootnode.transformation))

def _compute_bounding_box_for_node(nodes, node, bb_min, bb_max, transformation):
    
    if node.type == MESH:
        for v in node.aabb:
            v = transform(v, transformation)
            bb_min[0] = min(bb_min[0], v[0])
            bb_min[1] = min(bb_min[1], v[1])
            bb_min[2] = min(bb_min[2], v[2])
            bb_max[0] = max(bb_max[0], v[0])
            bb_max[1] = max(bb_max[1], v[1])
            bb_max[2] = max(bb_max[2], v[2])

    for child in node.children:
        bb_min, bb_max = _compute_bounding_box_for_node(nodes, nodes[child], bb_min, bb_max, transformation)

    return bb_min, bb_max

def get_bounding_box_for_node(scene, node):
    bb_min = [1e10, 1e10, 1e10] # x,y,z
    bb_max = [-1e10, -1e10, -1e10] # x,y,z

    global_transformation = get_world_transform(scene, node)

    return _compute_bounding_box_for_node(scene.nodes, node, bb_min, bb_max, global_transformation)

def get_world_transform(scene, node):

    if node == scene.rootnode:
        return numpy.identity(4, dtype=numpy.float32)

    parents = reversed(_get_parent_chain(scene, node, []))
    parent_transform = reduce(numpy.dot, [p.transformation for p in parents])
    return numpy.dot(parent_transform, node.transformation)


def _get_parent_chain(scene, node, parents):

    parent = scene.nodes[node.parent]

    parents.append(parent)

    if parent == scene.rootnode:
        return parents

    return _get_parent_chain(scene, parent, parents)

def transformed_aabb(scene, node):
    """ TODO: this computation is incorrect! To compute a transformed AABB, we need to
    re-compute the AABB from the transformed *vertices* (or maybe from the transformed bounding box?)
    """
    parents = reversed(_get_parent_chain(scene, node, []))

    global_transformation = reduce(numpy.dot, [p.transformation for p in parents])
    
    return transform(node.aabb[0], global_transformation), \
           transform(node.aabb[1], global_transformation)



import uuid
import copy
import json
import time

import logging
logger = logging.getLogger("underworlds.core")

import numpy

import underworlds.underworlds_pb2 as gRPC

from underworlds.errors import *

# Clients types
READER = gRPC.ClientInteraction.READER
PROVIDER = gRPC.ClientInteraction.PROVIDER
MONITOR = gRPC.ClientInteraction.MONITOR
FILTER = gRPC.ClientInteraction.FILTER

CLIENTTYPE_NAMES = {READER:'reader',
                    PROVIDER:'provider',
                    MONITOR:'monitor',
                    FILTER:'filter'
                   }

# Node types
UNDEFINED = gRPC.Node.UNDEFINED
# Entities are abstract nodes. They can represent non-physical objects (like a
# reference frame) or groups of other objects.
ENTITY = gRPC.Node.ENTITY
MESH = gRPC.Node.MESH
CAMERA = gRPC.Node.CAMERA

NODETYPE_NAMES = {UNDEFINED:'undefined',
                  ENTITY:'entity',
                  MESH:'mesh',
                  CAMERA:'camera'
                 }
# Situation types
GENERIC = gRPC.Situation.GENERIC
MOTION = gRPC.Situation.MOTION
EVT_MODELLOAD = gRPC.Situation.EVT_MODELLOAD

SITUATIONTYPE_NAMES = {GENERIC: "generic situation",
                       MOTION: "motion",
                       EVT_MODELLOAD: "model loading"
                      }

class Node(object):
    def __init__(self, name = "", type = UNDEFINED):

        ################################################################
        ##                     START OF THE API                       ##
        ################################################################
        self.id = str(uuid.uuid4())
        self.name = name
        self.type = type #one of the node constant defined in types.py
        self.parent = None # the parent node id
        #TODO: make children read-only with @property
        self.children = [] # the children nodes id. READ-ONLY
        

        # 4x4 transformation matrix, relative to parent. Stored as a numpy 4x4
        # matrix. Translation units are meters.
        self.transformation = numpy.identity(4, dtype=numpy.float32)

        self.last_update = time.time()

        self.lowres = []
        self.hires = []
        self.cad = []
        self.aabb = []

        self.properties = {
                "physics": False # no physics applied by default
            }

        self.clipplanenear = 0
        self.clipplanefar = 0
        self.aspect = 0
        self.horizontalfov = 0

        ################################################################
        ##                     END OF THE API                         ##
        ################################################################

    def __repr__(self):
        return self.id + (" (" + self.name + ")" if self.name else "")
    
    
    def __str__(self):
        return self.name if self.name else self.id + " (" + NODETYPE_NAMES[self.type] + ")"

    def __lt__(self, node):
        return self.id < node.id

    def __eq__(self, node):
        return hasattr(node, "id") and self.id == node.id

    def __hash__(self):
        return hash(self.id)

    def translate(self, vector):
        """ Translates the node by a vector. The change is not propagated
        automatically (you must call nodes.update(...)).
        """
        self.transformation[0,3] = vector[0]
        self.transformation[1,3] = vector[1]
        self.transformation[2,3] = vector[2]

    def copy(self):
        """ Performs a deep-copy of myself, and return the copy.
        The copy *has a new, different, unique ID* (ie, the ID is not copied).
        """

        import copy
        node = copy.deepcopy(self)
        node.id = str(uuid.uuid4())

        return node

    def serialize(self, NodeType):
        """Outputs a protobuf encoding of the node

        The NodeType (underworlds_pb2.Node) needs to be passed as parameter
        to prevent the creation of a 2nd instance of the underworlds_pb2 that
        crashes the gRPC. Not sure why...
        Similar to http://stackoverflow.com/questions/32010905/unbound-method-must-be-called-with-x-instance-as-first-argument-got-x-instance
        """


        node = NodeType()
        node.id = self.id
        node.name = self.name
        node.type = self.type
        node.parent = self.parent if self.parent is not None else ""

        for c in self.children:
            node.children.append(c)

        for v in self.transformation.flatten().tolist():
            node.transformation.append(v)

        node.last_update = self.last_update


        for m in self.lowres: node.lowres.append(m)
        for m in self.hires: node.hires.append(m)
        for m in self.cad: node.cad.append(m)

        for p in self.aabb:
            point = node.aabb.add()
            point.x, point.y, point.z = float(p[0]), float(p[1]), float(p[2])

        node.physics = self.properties["physics"]

        node.clipplanenear = self.clipplanenear
        node.clipplanefar =  self.clipplanefar
        node.aspect =        self.aspect
        node.horizontalfov = self.horizontalfov

        return node

    @staticmethod
    def deserialize(data):
        """Creates a node from a protobuf encoding.
        """
        node = Node()

        node.id = data.id
        node.name = data.name
        node.type = data.type
        node.parent = data.parent if data.parent else None # convert empty string to None if needed

        for c in data.children:
            node.children.append(c)

        node.transformation = [v for v in data.transformation]

        # Convert the transformation into a proper numpy array.
        # The type (float32) ensures OpenGL compatibilty on 64bit platforms
        node.transformation = numpy.array(node.transformation, dtype=numpy.float32).reshape(4,4)

        node.last_update = data.last_update


        node.lowres = [m for m in data.lowres]
        node.hires = [m for m in data.hires]
        node.cad = [m for m in data.cad]

        node.aabb = [(p.x,p.y,p.z) for p in data.aabb]
        if node.aabb and len(node.aabb) != 2:
            raise RuntimeError("Received a node with an invalid bounding-box!")

        node.properties["physics"] = data.physics

        node.clipplanenear = data.clipplanenear
        node.clipplanefar =  data.clipplanefar
        node.aspect =        data.aspect
        node.horizontalfov = data.horizontalfov

        return node

class Mesh(object):

    def __init__(self, vertices, faces, normals, diffuse=(1,1,1,1)):

        self.id = ""
        self.vertices = vertices
        self.faces = faces
        self.normals = normals
        self.diffuse = tuple(diffuse) # diffuse color, white by default
        if len(self.diffuse) == 3: self.diffuse += (1,)

        self.id = str(hash(str(self.serialize(gRPC.Mesh))))

    def __hash__(self):
        m = (self.vertices, \
             self.faces, \
             self.normals, \
             self.diffuse)
        return hash(str(m))

    def serialize(self, MeshType):
        """Outputs a protobuf encoding of the mesh

        The MeshType (underworlds_pb2.Mesh) needs to be passed as parameter
        to prevent the creation of a 2nd instance of the underworlds_pb2 that
        crashes the gRPC. Not sure why...
        Similar to http://stackoverflow.com/questions/32010905/unbound-method-must-be-called-with-x-instance-as-first-argument-got-x-instance
        """
        starttime = time.time()

        mesh = MeshType()
        mesh.id = self.id

        for vertex in self.vertices:
            point = mesh.vertices.add()
            point.x, point.y, point.z = vertex

        for f in self.faces:
            face = mesh.faces.add()
            face.x, face.y, face.z = f

        for normal in self.normals:
            point = mesh.normals.add()
            point.x, point.y, point.z = normal

        mesh.diffuse.r, mesh.diffuse.g, mesh.diffuse.b, mesh.diffuse.a = self.diffuse

        logger.info("Serialized mesh %s in %.2fsec" % (self.id, time.time()-starttime))
        return mesh


    @staticmethod
    def deserialize(data):
        """Creates a Python mesh object from a protobuf encoding.
        """

        mesh = Mesh(vertices=[(p.x,p.y,p.z) for p in data.vertices],
                    faces = [(f.x,f.y,f.z) for f in data.faces],
                    normals = [(n.x,n.y,n.z) for n in data.normals],
                    diffuse = (data.diffuse.r, data.diffuse.g, data.diffuse.b, data.diffuse.a))

        #if mesh.id != data.id:
        #    raise RuntimeError("Can not verify mesh integrity!")

        return mesh

class Scene(object):
    """An Underworlds scene
    """

    def __init__(self):

        self.rootnode = Node("root", ENTITY)
        self.rootnode.transformation = numpy.identity(4, dtype=numpy.float32)

        self.nodes = []

        self.nodes.append(self.rootnode)

    def list_entities(self):
        """ Returns the list of entities contained in the scene.
        """
        raise NotImplementedError

    def node(self, id):
        """ Returns a node from its ID (or None if the node does not exist)
        """
        for n in self.nodes:
            if n.id == id:
                return n

    def nodebyname(self, name):
        """ Returns a list of node that have the given name (or [] if no node has this name)
        """
        nodes = []
        for n in self.nodes:
            if n.name == name:
                nodes.append(n)
        return nodes

class Timeline(object):
    """ Stores 'situations' (ie, either events -- temporal objects
    without duration -- or static situations -- temporal objects
    with a non-null duration).

    A timeline also exposes an API to find for temporal patterns.

    TODO: situations are currently stored as a flat array, which
    is certainly not the most efficient way!
    """

    def __init__(self):

        self.origin = time.time()

        self.situations = {}

    def on(self, event):
        """
        Creates a new EventMonitor to watch a given event model.

        Typical use is:
        
        >>> t = Timeline()
        >>> e = Event(...)
        >>>
        >>> def onevt(evt):
        >>>    print(evt)
        >>>
        >>> t.on(e).call(onevt)

        :returns: a new instance of EventMonitor for this event.
        """
        return EventMonitor(event)


    def start(self):
        """ Asserts a situation has started.

        Note that in the special case of events, the situation ends
        immediately.

        :returns: The newly created situation
        """
        situation = Situation()
        situation.starttime = time.time()
        self.situations[situation.id] = situation
        return situation

    def end(self, situation):
        """ Asserts the end of a situation.

        Note that in the special case of events, this method a no effect.
        """
        if situation.isevent():
            return situation
        self.situations[situation.id].endtime = time.time()
        return self.situations[situation.id]

    def event(self):
        """ Asserts a new event occured in this timeline.

        :returns: the newly created event.
        """
        event = self.start()
        event.endtime = event.starttime
        return event

    def append(self, situation):
        """ Adds the given situation to the timeline.

        If a situation with the same ID already exists, it replaces it.

        :returns: True if the situation has been added, False if it has simply updated an existing situation
        """
        return not self.update(situation)

    def update(self, situation):
        """ Update (ie, replace) an existing situation with the
        given one.
        
        If the situations does not exist, simply add it to the timeline.

        :returns: True if the situation has been updated, False if it has been simply added (new situation)
        """
        isnew = situation.id in self.situations
        self.situations[situation.id] = situation
        return not isnew

    def remove(self, situation):
        """ Deletes an existing situation.
        """
        del self.situations[situation.id]

    def situation(self, id):
        return self.situations[id]


class EventMonitor(object):

    def __init__(self, evt):
        self.evt = evt

    def call(self, cb):
        self.cb = cb

    def make_call(self):
        self.cb(self.evt)

    def wait(self, timeout = 0):
        """ Blocks until an event occurs, or the timeout expires.
        """
        raise NotImplementedError

class World(object):

    def __init__(self, name):

        self.name = name
        self.scene = Scene()
        self.timeline = Timeline()

    def __repr__(self):
        return "world " + self.name

    def deepcopy(self, world):
        self.scene = copy.copy(world.scene)
        self.timeline = copy.copy(world.timeline)


class Situation(object):
    """ A situation represents a generic temporal object.

    It has two subclasses:
     - events, which are instantaneous situations (null duration)
     - static situations, that have a duration.

     """

    def __init__(self, desc="", type = GENERIC):

        self.id = str(uuid.uuid4())
        self.type = type
        self.desc = desc

        self.last_update = time.time()

        # Start|Endtime are in seconds (float)
        self.starttime = 0 # convention for situations that are not yet started
        self.endtime = 0 # convention for situations that are not terminated

    def isevent(self):
        return self.endtime == self.starttime

    def __repr__(self):
        return self.id + " (" + SITUATIONTYPE_NAMES[self.type] + ")"

    def __str__(self):
        if self.desc:
            return "Situation %s -- %s: %s" % (self.id, SITUATIONTYPE_NAMES[self.type], self.desc)
        else:
            return "Situation %s -- %s" % (self.id, SITUATIONTYPE_NAMES[self.type])

    def __cmp__(self, sit):
        # TODO: check here other values equality, and raise exception if any differ? may be costly, though...
         return cmp(self.id, sit.id)

    def __hash__(self):
        return hash(self.id)

    def serialize(self, SituationType):
        """Outputs a protobuf encoding of the situation

        The SituationType (underworlds_pb2.Situation) needs to be passed as parameter
        to prevent the creation of a 2nd instance of the underworlds_pb2 that
        crashes the gRPC. Not sure why...
        Similar to http://stackoverflow.com/questions/32010905/unbound-method-must-be-called-with-x-instance-as-first-argument-got-x-instance
        """
        sit = gRPC.Situation()

        sit.id = self.id
        sit.type = self.type
        sit.description = self.desc
        sit.last_update = self.last_update
        sit.start.time = self.starttime
        sit.end.time = self.endtime


        return sit


    @staticmethod
    def deserialize(data):
        """Creates a situation from a protobuf encoding.
        """
        sit = Situation()

        sit.id = data.id
        sit.type = data.type
        sit.desc = data.description
        sit.last_update = data.last_update
        sit.starttime = data.start.time
        sit.endtime = data.end.time

        return sit





import OpenGL
#OpenGL.ERROR_CHECKING=False
#OpenGL.ERROR_LOGGING = False
#OpenGL.ERROR_ON_COPY = True
#OpenGL.FULL_LOGGING = True
from OpenGL.GL import *
from OpenGL.error import GLError
from OpenGL.GLU import *
from OpenGL.arrays import vbo
from OpenGL.GL import shaders

import time
import math, random
import numpy
from numpy import linalg

import logging
logger = logging.getLogger("underworlds.visibility")

import underworlds
from underworlds.types import *
from underworlds.helpers.geometry import transform, get_world_transform
from underworlds.helpers import transformations


FLAT_VERTEX_SHADER="""
#version 130

uniform mat4 u_viewProjectionMatrix;
uniform mat4 u_modelMatrix;

uniform vec4 u_materialDiffuse;

in vec3 a_vertex;

out vec4 v_color;

void main(void)
{
    v_color = u_materialDiffuse;
    gl_Position = u_viewProjectionMatrix * u_modelMatrix * vec4(a_vertex, 1.0);
}
"""

BASIC_FRAGMENT_SHADER="""
#version 130

in vec4 v_color;

void main() {
    gl_FragColor = v_color;
}
"""
ROTATION_180_X = numpy.array([[1,0,0,0],[0,-1,0,0],[0,0,-1,0],[0,0,0,1]], dtype=numpy.float32)
DEFAULT_CLIP_PLANE_NEAR = 0.001
DEFAULT_CLIP_PLANE_FAR = 1000.0

class VisibilityMonitor:


    def __init__(self, ctx, world, w=80, h=60, create_surface=True, debug=False):

        self.debug = debug

        self.w = w
        self.h = h

        if create_surface:
            import pygame
            pygame.init()
            if not debug:
                pygame.display.set_mode((w,h), pygame.OPENGL | pygame.DOUBLEBUF)
                pygame.display.iconify()
            else:
                pygame.display.set_mode((w,h), pygame.OPENGL | pygame.DOUBLEBUF | pygame.RESIZABLE)

        self.prepare_shaders()

        self.ctx = ctx
        self.world = world

        self.scene = None
        self.meshes = {} # stores the OpenGL vertex/faces/normals buffers pointers

        self.node2colorid = {} # stores a color ID for each node. Useful for mouse picking and visibility checking
        self.colorid2node = {} # reverse dict of node2colorid

        self.cameras = []

        self.load_world()

        if not self.cameras:
            raise RuntimeError("No camera in the world <%s>. Giving up." % self.world)

    def prepare_shaders(self):

        ### Flat shader
        flatvertex = shaders.compileShader(FLAT_VERTEX_SHADER, GL_VERTEX_SHADER)
        fragment = shaders.compileShader(BASIC_FRAGMENT_SHADER, GL_FRAGMENT_SHADER)

        self.flatshader = shaders.compileProgram(flatvertex,fragment)

        self.set_shader_accessors( ('u_modelMatrix',
                                    'u_viewProjectionMatrix',
                                    'u_materialDiffuse',), 
                                    ('a_vertex',), self.flatshader)


    def set_shader_accessors(self, uniforms, attributes, shader):
        # add accessors to the shaders uniforms and attributes
        for uniform in uniforms:
            location = glGetUniformLocation( shader,  uniform )
            if location in (None,-1):
                raise RuntimeError('No uniform: %s (maybe it is not used '
                                   'anymore and has been optimized out by'
                                   ' the shader compiler)'%( uniform ))
            setattr( shader, uniform, location )

        for attribute in attributes:
            location = glGetAttribLocation( shader, attribute )
            if location in (None,-1):
                raise RuntimeError('No attribute: %s'%( attribute ))
            setattr( shader, attribute, location )

    def prepare_gl_buffers(self, id):

        meshes = self.meshes

        if id in meshes:
            # mesh already loaded. Fine
            return

        meshes[id] = {}

        # leave some time for new nodes to push their meshes
        if not self.ctx.has_mesh(id):
            logger.warning("Mesh ID %s is not available on the server... "
                           "waiting for it..." % id)
            while not self.ctx.has_mesh(id):
                time.sleep(0.01)

            logger.info("Mesh ID %s is now available. Getting it..." % id)

        mesh = self.ctx.mesh(id) # retrieve the mesh from the server

        # Fill the buffer for vertex
        v = numpy.array(mesh.vertices, 'f')

        meshes[id]["vbo"] = vbo.VBO(v)

        # Fill the buffer with faces indices
        meshes[id]["faces"] = glGenBuffers(1)
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, meshes[id]["faces"])
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, 
                    numpy.array(mesh.faces, dtype=numpy.int32),
                    GL_STATIC_DRAW)

        meshes[id]["nbfaces"] = len(mesh.faces)

        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,0)

    def get_rgb_from_colorid(self, colorid):
        r = (colorid >> 0) & 0xff
        g = (colorid >> 8) & 0xff
        b = (colorid >> 16) & 0xff

        return (r,g,b)

    def get_color_id(self):
        id = random.randint(0, 256*256*256)
        if id not in self.colorid2node:
            return id
        else:
            return self.get_color_id()

    def glize(self, node):


        if node.type == MESH:
            colorid = self.get_color_id()
            self.colorid2node[colorid] = node
            self.node2colorid[node] = colorid

            if hasattr(node, "cad"):
                node.glmeshes = node.cad
            elif hasattr(node, "lowres"):
                node.glmeshes = node.lowres
            elif hasattr(node, "hires"):
                node.glmeshes = node.hires
            else:
                raise StandardError("The node %s has no mesh available!" % node.name)
            for mesh in node.glmeshes:
                self.prepare_gl_buffers(mesh)

        elif node.type == CAMERA:
            logger.info("Added camera <%s>" % node.name)
            self.cameras.append(node)


    def load_world(self):
        logger.info("Preparing world <%s> for 3D rendering..." % self.world)

        scene = self.scene = self.world.scene
        nodes = scene.nodes
        for node in nodes:
            logger.info("Loading node <%s>" % node)
            self.glize(node)

        logger.info("World <%s> ready for visibility monitoring." % self.world)

    def set_camera(self, name):

        camera = None
        for c in self.cameras:
            if c.name == name:
                camera = c
                break

        if camera is None:
            raise RuntimeError("Camera <%s> does not exist in world <%s>" % (name, self.world.name))

        # Update the camera position from the server
        camera = self.scene.nodes[camera.id]

        znear = camera.clipplanenear or DEFAULT_CLIP_PLANE_NEAR
        zfar = camera.clipplanefar or DEFAULT_CLIP_PLANE_FAR
        aspect = camera.aspect
        fov = camera.horizontalfov

        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()

        # Compute gl frustrum
        tangent = math.tan(fov/2.)
        h = znear * tangent
        w = h * aspect

        # params: left, right, bottom, top, near, far
        glFrustum(-w, w, -h, h, znear, zfar)
        # equivalent to:
        #gluPerspective(fov * 180/math.pi, aspect, znear, zfar)

        self.projection_matrix = glGetFloatv( GL_PROJECTION_MATRIX).transpose()

        self.view_matrix = linalg.inv(get_world_transform(self.scene,camera))

        # Rotate by 180deg around X to have Z pointing backward (OpenGL convention)
        self.view_matrix = numpy.dot(ROTATION_180_X, self.view_matrix)

        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()
        glMultMatrixf(self.view_matrix.transpose())


    def render_colors(self):

        glEnable(GL_DEPTH_TEST)
        glDepthFunc(GL_LEQUAL)

        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
        glEnable(GL_CULL_FACE)

        glUseProgram(self.flatshader)

        glUniformMatrix4fv( self.flatshader.u_viewProjectionMatrix, 1, GL_TRUE,
                            numpy.dot(self.projection_matrix,self.view_matrix))


        self.recursive_render(self.scene.rootnode, self.flatshader)

        glUseProgram( 0 )

    def compute_all(self):
        """
        :returns: dictionary {camera: [visible nodes]}
        Attention: The performances of this method relies heavily on the size of the display!
        """
        visible_objects = {}

        for c in self.cameras:
            visible_objects[c.name] = self.from_camera(c.name)

        return visible_objects

    def from_camera(self, camera):
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        self.set_camera(camera)
        self.render_colors()
        # Capture image from the OpenGL buffer
        buf = ( GLubyte * (3 * self.w * self.h) )(0)
        glReadPixels(0, 0, self.w, self.h, GL_RGB, GL_UNSIGNED_BYTE, buf)

        #Reinterpret the RGB pixel buffer as a 1-D array of 24bits colors
        a = numpy.ndarray(len(buf), numpy.dtype('>u1'), buf)
        colors = numpy.zeros(int(len(buf) / 3), numpy.dtype('<u4'))
        for i in range(3):
            colors.view(dtype='>u1')[i::4] = a.view(dtype='>u1')[i::3]

        seen = numpy.unique(colors)
        seen.sort()
        seen = seen[1:] if seen[0] == 0 else seen # remove the 0 for background

        if self.debug:
            print("World seen from camera %s" % camera)
            import pygame
            pygame.display.flip()
            time.sleep(1)

        return [self.colorid2node[i] for i in seen]

        #colors = colors[numpy.nonzero(colors)] #discard black background
        
        #if colors.any():
        #    bins = numpy.bincount(colors)
        #    ii = numpy.nonzero(bins)[0]

        #    for i in ii:
        #        print ("Node %s is visible (%d pix)" % (self.colorid2node[i], bins[i]))
        #else:
        #    print("Nothing visible!")


    def recursive_render(self, node, shader):
        """ Main recursive rendering method.
        """

        try:
            m = get_world_transform(self.scene, node)
        except AttributeError:
            #probably a new incoming node, that has not yet been converted to numpy
            self.glize(node)
            m = get_world_transform(self.scene, node)

        if node.type == MESH:

            # if the node has been recently turned into a mesh, we might not
            # have the mesh data yet.
            if not hasattr(node, "glmeshes"):
                self.glize(node)

            for id in node.glmeshes:

                stride = 12 # 3 * 4 bytes

                colorid = self.node2colorid[node]
                r,g,b= self.get_rgb_from_colorid(colorid)
                glUniform4f( shader.u_materialDiffuse, r/255.0,g/255.0,b/255.0,1.0 )

                glUniformMatrix4fv( shader.u_modelMatrix, 1, GL_TRUE, m )

                vbo = self.meshes[id]["vbo"]
                vbo.bind()

                glEnableVertexAttribArray( shader.a_vertex )

                glVertexAttribPointer(
                    shader.a_vertex,
                    3, GL_FLOAT,False, stride, vbo
                )

                glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self.meshes[id]["faces"])
                glDrawElements(GL_TRIANGLES, self.meshes[id]["nbfaces"] * 3, GL_UNSIGNED_INT, None)


                vbo.unbind()
                glDisableVertexAttribArray( shader.a_vertex )

                glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0)

        for child in node.children:
            try:
                self.recursive_render(self.scene.nodes[child], shader)
            except KeyError as ke:
                logger.warning("Node ID %s listed as child of %s, but it"
                                " does not exist! Skipping it" % (child, repr(node)))







import os # for the UWDS_SERVER environment variable
import sys

import time
import copy
import threading
import random

from collections import deque

import logging
logger = logging.getLogger("underworlds.client")

from grpc.beta import implementations
from grpc.framework.interfaces.face.face import ExpirationError,NetworkError,AbortionError
import underworlds.underworlds_pb2 as gRPC

from underworlds.types import World, Node, Situation, Mesh

from underworlds.helpers.profile import profile, profileonce

_TIMEOUT_SECONDS = 1
_TIMEOUT_SECONDS_MESH_LOADING = 20

#TODO: inherit for a collections.MutableSequence? what is the benefit?
class NodesProxy:

    def __init__(self, context, world):

        self._ctx = context # current underworlds context (useful to know the client name)
        self._world = world

        # This contains the tuple (id, world) and is used for identification
        # when communicating with the server
        self._server_ctx = gRPC.Context(client=self._ctx.id, world=self._world.name)

        self._len = self._ctx.rpc.getNodesLen(self._server_ctx, _TIMEOUT_SECONDS).size

        self._nodes = {} # node store

        # list of all node IDs that were once obtained.
        # They may be valid or invalid (if present in _updated_ids)
        self._ids = []

        # list of invalid ids (ie, nodes that have remotely changed).
        # This list is updated asynchronously from a server publisher
        self._updated_ids = deque(self._ctx.rpc.getNodesIds(self._server_ctx, _TIMEOUT_SECONDS).ids)

        self._deleted_ids = deque()

        # holds futures for non-blocking RPC calls when updating/removing nodes
        self.update_future = None
        self.remove_future = None

        # Get the root node
        self.rootnode = self._ctx.rpc.getRootNode(self._server_ctx, _TIMEOUT_SECONDS).id
        self._update_node_from_remote(self.rootnode)
 
        self.waitforchanges_cv = threading.Condition()
        self.lastchange = None

    @profile
    def _on_remotely_updated_node(self, id):

        if id not in self._updated_ids:
            self._updated_ids.append(id)

        with self.waitforchanges_cv:
            self.lastchange = (id, gRPC.Invalidation.UPDATE)
            self.waitforchanges_cv.notify_all()


    @profile
    def _on_remotely_added_node(self, id):

        self._len += 1 # not atomic, but still fine since I'm the only one to write it

        if id not in self._updated_ids:
            self._updated_ids.append(id)

        with self.waitforchanges_cv:
            self.lastchange = (id, gRPC.Invalidation.NEW)
            self.waitforchanges_cv.notify_all()


    @profile
    def _on_remotely_deleted_node(self, id):

        self._len -= 1 # not atomic, but still fine since I'm the only one to write it
        self._deleted_ids.append(id)

        with self.waitforchanges_cv:
            self.lastchange = (id, gRPC.Invalidation.DELETE)
            self.waitforchanges_cv.notify_all()


    def _get_more_node(self):
        
        if not self._updated_ids:
            logger.warning("Slow propagation? Waiting for new/updated nodes notifications...")
            time.sleep(0.05) #leave some time for propagation

            # still empty? we have a problem!
            if not self._updated_ids:
                logger.error("Inconsistency detected! The server has not"\
                             " notified all the nodes updates. Or the "\
                             "IPC transport is really slow.")
                raise Exception()

        # here, _updated_ids is not empty. It should not raise an exception
        id = self._updated_ids.pop()

        self._get_node_from_remote(id)

    def _get_node_from_remote(self, id):

        nodeInCtxt = gRPC.NodeInContext(context=self._server_ctx,
                                        node=gRPC.Node(id=id))
        try:
            gRPCNode = self._ctx.rpc.getNode(nodeInCtxt, _TIMEOUT_SECONDS)
        except AbortionError as e:
            raise IndexError(e.details)

        # is it a new node, or rather an update to an existing one?
        if id not in self._ids:
            self._ids.append(id)

        self._nodes[id] = Node.deserialize(gRPCNode)


    def _update_node_from_remote(self, id):

        self._get_node_from_remote(id)

        self._updated_ids.remove(id)

    def __getitem__(self, key):

        # First, a bit of house keeping
        # do we have pending nodes to delete?
        if self._deleted_ids:
            tmp = copy.copy(self._deleted_ids)
            for id in tmp:
                try:
                    self._ids.remove(id)
                    del(self._nodes[id])
                    self._deleted_ids.remove(id)
                except ValueError:
                    logger.warning("The node %s is already removed. Feels like a synchro issue..." % id)

        # Then, let see what the user want:
        if type(key) is int:

            # First, are we over the length of our node list?
            if key >= self._len:
                raise IndexError

            # not downloaded enough nodes yet?
            while key >= len(self._ids):
                self._get_more_node()

            id = self._ids[key]

            # did the node changed since the last time we obtained it?
            if id in self._updated_ids:
                self._update_node_from_remote(id)

            return self._nodes[id]

        else: #assume it's a node ID

            if key in self._ids:
                # did the node changed since the last time we obtained it?
                if key in self._updated_ids:
                        self._update_node_from_remote(key)
                return self._nodes[key]
            elif key in self._updated_ids:
                self._update_node_from_remote(key)
                return self._nodes[key]
            else: # we do not have this node locally. Let's try to fetch it
                try:
                    self._get_node_from_remote(key)
                except ValueError:
                    #The node does not exist!!
                    raise KeyError("The node ID %s does not exist" % key)
                return self._nodes[key]

    def __len__(self):
        return self._len

    def append(self, node):
        """ Adds a new node to the node set.

        It is actually an alias for NodesProxy.update: all the restrictions
        regarding ordering or propagation time apply as well.
        """
        return self.update(node)

    @profile
    def update(self, node):
        """ Update the value of a node in the node set.
        If the node does not exist yet, add it.

        This method sends the new/updated node to the
        remote. IT DOES NOT DIRECTLY modify the local
        copy of nodes: the roundtrip is slower, but data
        consistency is easier to ensure.

        This means that if you create or update a node, the
        node won't be created/updated immediately. It will 
        take some time (a couple of milliseconds) to propagate
        the change.

        Also, you have no guarantee regarding the ordering:

        for instance,

        >>> nodes.update(n1)
        >>> nodes.update(n2)

        does not mean that nodes[0] = n1 and nodes[1] = n2.
        This is due to the lazy access process.

        However, once accessed once, nodes keep their index (until a
        node with a smaller index is removed).

        """

        # if for some reason, the previous non-blocking call to update the node is
        # not yet complete, finish it now
        if self.update_future is not None:
            self.update_future.result()

        self.update_future = self._ctx.rpc.updateNode.future(
                                 gRPC.NodeInContext(context=self._server_ctx,
                                                    node=node.serialize(gRPC.Node)),
                                 _TIMEOUT_SECONDS)

    @profile
    def remove(self, node):
        """ Deletes a node from the node set.

        THIS METHOD DOES NOT DIRECTLY delete the local
        copy of the node: it tells instead the server to
        delete this node for all clients.
        the roundtrip is slower, but data consistency is easier to ensure.

        This means that if you delete a node, the
        node won't be actually deleted immediately. It will 
        take some time (a couple of milliseconds) to propagate
        the change.
        """
        # if for some reason, the previous non-blocking call to delete the node is
        # not yet complete, finish it now
        if self.remove_future is not None:
            self.remove_future.result()

        self.remove_future = self._ctx.rpc.deleteNode.future(
                                 gRPC.NodeInContext(context=self._server_ctx,
                                                    node=node.serialize(gRPC.Node)),
                                 _TIMEOUT_SECONDS)


class SceneProxy(object):

    def __init__(self, ctx, world):

        self._ctx = ctx # context

        self.nodes = NodesProxy(self._ctx, world)

    @property
    def rootnode(self):
        return self.nodes[self.nodes.rootnode]

    def waitforchanges(self, timeout = None):
        """ This method blocks until either the scene has
        been updated (a node has been either updated, 
        added or removed) or the timeout is over.

        :param timeout: timeout in seconds (float value)
        :returns: the change that occured as a pair [node id, operation]
        (operation is one of gRPC.Invalidation.UPDATE,
        gRPC.Invalidation.NEW, gRPC.Invalidation.DELETE) or None if the
        timeout has been reached.
        """
        lastchange = None
        with self.nodes.waitforchanges_cv:
            self.nodes.waitforchanges_cv.wait(timeout)
            lastchange = self.nodes.lastchange
            self.nodes.lastchange = None

        profileonce("client.scene.waitforchanges notified")

        return lastchange


    def nodebyname(self, name):
        """ Returns a list of node that have the given name (or [] if no node has this name)
        """
        nodes = []
        for n in self.nodes:
            if n.name == name:
                nodes.append(n)
        return nodes

    def append_and_propagate(self, node):
        """An alias for NodesProxy.append
        """
        self.nodes.append(node)

    def update_and_propagate(self, node):
        """An alias for NodesProxy.update
        """
        self.nodes.update(node)

    def remove_and_propagate(self, node):
        """An alias for NodesProxy.remove
        """
        self.nodes.remove(node)



class TimelineProxy:

    def __init__(self, ctx, world):

        self._ctx = ctx # context
        self._world = world

        # This contains the tuple (id, world) and is used for identification
        # when communicating with the server
        self._server_ctx = gRPC.Context(client=self._ctx.id, world=self._world.name)

        self._len = self._ctx.rpc.getSituationsLen(self._server_ctx, _TIMEOUT_SECONDS).size

        self._situations = {}

        # list of all node IDs that were once obtained.
        # They may be valid or invalid (if present in _updated_ids)
        self._ids = []

        # list of invalid ids (ie, nodes that have remotely changed).
        # This list is updated asynchronously from a server publisher
        self._updated_ids = deque(self._ctx.rpc.getSituationsIds(self._server_ctx, _TIMEOUT_SECONDS).ids)

        self._deleted_ids = deque()

        # holds futures for non-blocking RPC calls when updating/removing nodes
        self.update_future = None
        self.remove_future = None


        self.origin = self._ctx.rpc.timelineOrigin(self._server_ctx, _TIMEOUT_SECONDS).time
        logger.info("Accessing world <%s> (initially created on  %s)"%(self._world.name, time.asctime(time.localtime(self.origin))))

        self.waitforchanges_cv = threading.Condition()
        self.lastchange = None

    @profile
    def _on_remotely_added_situation(self, id):


        self._len += 1 # not atomic, but still fine since I'm the only one to write it

        if id not in self._updated_ids:
            self._updated_ids.append(id)

        with self.waitforchanges_cv:
            self.lastchange = (id, gRPC.Invalidation.NEW)
            self.waitforchanges_cv.notify_all()

    @profile
    def _on_remotely_updated_situation(self, id):

        if id not in self._updated_ids:
            self._updated_ids.append(id)

        with self.waitforchanges_cv:
            self.lastchange = (id, gRPC.Invalidation.UPDATE)
            self.waitforchanges_cv.notify_all()


    @profile
    def _on_remotely_deleted_situation(self, id):

        self._len -= 1 # not atomic, but still fine since I'm the only one to write it
        self._deleted_ids.append(id)

        with self.waitforchanges_cv:
            self.lastchange = (id, gRPC.Invalidation.DELETE)
            self.waitforchanges_cv.notify_all()

    def _get_more_situations(self):
        
        if not self._updated_ids:
            logger.warning("Slow propagation? Waiting for new/updated situations notifications...")
            time.sleep(0.05) #leave some time for propagation

            # still empty? we have a problem!
            if not self._updated_ids:
                logger.error("Inconsistency detected! The server has not"\
                             " notified all the situations updates. Or the "\
                             "IPC transport is really slow.")
                raise Exception()

        # here, _updated_ids is not empty. It should not raise an exception
        id = self._updated_ids.pop()

        self._get_situation_from_remote(id)

    def _get_situation_from_remote(self, id):

        sitInCtxt = gRPC.SituationInContext(context=self._server_ctx,
                                        situation=gRPC.Situation(id=id))
        try:
            gRPCSituation = self._ctx.rpc.getSituation(sitInCtxt, _TIMEOUT_SECONDS)
        except AbortionError as e:
            raise IndexError(e.details)

        # is it a new situation, or rather an update to an existing one?
        if id not in self._ids:
            self._ids.append(id)

        self._situations[id] = Situation.deserialize(gRPCSituation)

    def _update_situation_from_remote(self, id):

        self._get_situation_from_remote(id)

        self._updated_ids.remove(id)

    def __contains__(self, situation):
        try:
            self[situation.id]
            return True
        except KeyError:
            return False
        

    def __getitem__(self, key):

        # First, a bit of house keeping
        # do we have pending situations to delete?
        if self._deleted_ids:
            tmp = copy.copy(self._deleted_ids)
            for id in tmp:
                try:
                    self._ids.remove(id)
                    del(self._situations[id])
                    self._deleted_ids.remove(id)
                except ValueError:
                    logger.warning("The situation %s is already removed. Feels like a synchro issue..." % id)

        # Then, let see what the user want:
        if type(key) is int:

            # First, are we over the length of our situation list?
            if key >= self._len:
                raise IndexError

            # not downloaded enough nodes yet?
            while key >= len(self._ids):
                self._get_more_situations()

            id = self._ids[key]

            # did the situation changed since the last time we obtained it?
            if id in self._updated_ids:
                self._update_situation_from_remote(id)

            return self._situations[id]

        else: #assume it's a situation ID

            if key in self._ids:
                # did the situation changed since the last time we obtained it?
                if key in self._updated_ids:
                        self._update_situation_from_remote(key)
                return self._situations[key]
            elif key in self._updated_ids:
                self._update_situation_from_remote(key)
                return self._situations[key]
            else: # we do not have this situation locally. Let's try to fetch it
                try:
                    self._get_situation_from_remote(key)
                except ValueError:
                    #The situation does not exist!!
                    raise KeyError("The situation ID %s does not exist" % key)
                return self._situations[key]

    def __len__(self):
        return self._len

    def __repr__(self):
        return "TimelineProxy %s" % [str(s) for s in self._situations]

    def start(self):
        """ Starts (and return) a new situation

        This method sends the new situation to the
        remote. IT DOES NOT DIRECTLY modify the local
        copy of situations: the roundtrip is slower, but data
        consistency is easier to ensure.

        Note that the situation starting time is set by the client, at the time
        at which this method is called.
        """
        situation = Situation()
        situation.starttime = time.time()
        self.update(situation)
        return situation

    def end(self, situation):
        """ Ends an existing situation

        This method sends the situation update to the
        remote. IT DOES NOT DIRECTLY modify the local
        copy of situations: the roundtrip is slower, but data
        consistency is easier to ensure.

        Note that the situation ending time is set by the client, at the time
        at which this method is called.
        """
        if situation.isevent():
            return situation
        situation.endtime = time.time()
        self.update(situation)
        return situation

    def event(self):
        """ Creates (and return) a new event (situation with 0 duration)

        This method sends the new event to the
        remote. IT DOES NOT DIRECTLY modify the local
        copy of situations: the roundtrip is slower, but data
        consistency is easier to ensure.

        Note that the event time is set by the client, at the time
        at which this method is called.
        """
        situation = Situation()
        situation.starttime = time.time()
        situation.endtime = situation.starttime
        self.update(situation)
        return situation

    def append(self, situation):
        """ Adds a new situation to the timeline.

        It is actually an alias for TimelineProxy.update: all the restrictions
        regarding ordering or propagation time apply as well.
        """
        return self.update(situation)


    @profile
    def update(self, situation):
        """ Update the value of a situation.
        If the situation does not exist yet, add it.

        This method sends the new/updated situation to the
        remote. IT DOES NOT DIRECTLY modify the local
        copy of situations: the roundtrip is slower, but data
        consistency is easier to ensure.

        This means that if you create or update a situation, the
        situation won't be created/updated immediately. It will 
        take some time (a couple of milliseconds) to propagate
        the change.

        Also, you have no guarantee regarding the ordering:

        for instance,

        >>> timeline.update(sit1)
        >>> timeline.update(sit2)

        does not mean that timeline[0] = sit1 and timeline[1] = sit2.
        This is due to the lazy access process.

        However, once accessed once, situations keep their index (until a
        situation with a smaller index is removed).
        """

        # if for some reason, the previous non-blocking call to update the situation is
        # not yet complete, finish it now
        if self.update_future is not None:
            self.update_future.result()

        self.update_future = self._ctx.rpc.updateSituation.future(
                                 gRPC.SituationInContext(context=self._server_ctx,
                                                    situation=situation.serialize(gRPC.Situation)),
                                 _TIMEOUT_SECONDS)
        return situation

    @profile
    def remove(self, situation):
        """ Deletes a situation.

        THIS METHOD DOES NOT DIRECTLY delete the local
        copy of the situation: it tells instead the server to
        delete this situation for all clients.
        the roundtrip is slower, but data consistency is easier to ensure.

        This means that if you delete a situation, the
        situation won't be actually deleted immediately. It will 
        take some time (a couple of milliseconds) to propagate
        the change.
        """
        # if for some reason, the previous non-blocking call to delete the situation is
        # not yet complete, finish it now
        if self.remove_future is not None:
            self.remove_future.result()

        self.remove_future = self._ctx.rpc.deleteSituation.future(
                                 gRPC.SituationInContext(context=self._server_ctx,
                                                        situation=situation.serialize(gRPC.Situation)),
                                 _TIMEOUT_SECONDS)


    def waitforchanges(self, timeout = None):
        """ This method blocks until either the timeline has
        been updated (a situation has been either started or
        ended) or the timeout is over.

        :param timeout: timeout in seconds (float value)

        :returns: the change that occured as a pair [node id, operation]
        (operation is one of gRPC.Invalidation.UPDATE,
        gRPC.Invalidation.NEW, gRPC.Invalidation.DELETE) or None if the
        timeout has been reached.
        """
        lastchange = None
        with self.waitforchanges_cv:
            self.waitforchanges_cv.wait(timeout)
            lastchange = self.lastchange
            self.lastchange = None

        profileonce("client.timeline.waitforchanges notified")

        return lastchange



class WorldProxy:

    def __init__(self, ctx, name):
    
        if not (isinstance(name, str) or isinstance(name, unicode)):
            raise TypeError("A world proxy must be initialized "
                            "with a string as name. Got %s instead." % type(name))

        self._ctx = ctx # context

        self._world = World(name)

        self.name = name
        self.scene = SceneProxy(self._ctx, self._world)
        self.timeline = TimelineProxy(self._ctx, self._world)

    def copy_from(self, world):
        """ Creates and/or replaces the content of the world with an exact copy
        of the given `world`.
        """
        req = {"client":self._ctx.id,
               "world": self._world.name,
               "req": "deepcopy %s" % (world.name)}

        self._ctx.rpc.send_json(req)
        self._ctx.rpc.recv() #ack

    def __str__(self):
        return self.name

class WorldsProxy:

    def __init__(self, ctx):

        self._ctx = ctx # context

        self._worlds = []

    def __getitem__(self, key):
        for world in self._worlds:
            if world.name == key:
                return world

        world = WorldProxy(self._ctx, key)
        self._worlds.append(world)
        return world

    def __setitem__(self, key, world):
        logger.error("Can not set a world")
        pass

    def __iter__(self):
        """To iterate over the existing worlds, first ask the server
        an up-to-date list of worlds, and yield worlds as much as needed.
        Doing so, worlds are lazily created.
        """
        topo = self._ctx.rpc.topology(gRPC.Client(id=self._ctx.id), _TIMEOUT_SECONDS)
        for world in topo.worlds:
            yield self.__getitem__(world)

class InvalidationServer(gRPC.BetaUnderworldsInvalidationServicer):

    def __init__(self, ctx):
        self.ctx=ctx

    @profile
    def emitInvalidation(self, invalidation, context):
        logger.info("Got <emitInvalidation> for world <%s>" % invalidation.world)
       
        target, action, world, id = invalidation.target, invalidation.type, invalidation.world, invalidation.id

        if target == gRPC.Invalidation.SCENE:
            if action == gRPC.Invalidation.UPDATE:
                logger.debug("Server notification: node updated: " + id)
                self.ctx.worlds[world].scene.nodes._on_remotely_updated_node(id)
            elif action == gRPC.Invalidation.NEW:
                logger.debug("Server notification: node added: " + id)
                self.ctx.worlds[world].scene.nodes._on_remotely_added_node(id)
            elif action == gRPC.Invalidation.DELETE:
                logger.debug("Server notification: node deleted: " + id)
                self.ctx.worlds[world].scene.nodes._on_remotely_deleted_node(id)
            else:
                raise RuntimeError("Unexpected invalidation action")

        elif target == gRPC.Invalidation.TIMELINE:
            if action == gRPC.Invalidation.UPDATE:
                logger.debug("Server notification: situation updated: " + id)
                self.ctx.worlds[world].timeline._on_remotely_updated_situation(id)
            elif action == gRPC.Invalidation.NEW:
                logger.debug("Server notification: situation added: " + id)
                self.ctx.worlds[world].timeline._on_remotely_added_situation(id)
            elif action == gRPC.Invalidation.DELETE:
                logger.debug("Server notification: situation deleted: " + id)
                self.ctx.worlds[world].timeline._on_remotely_deleted_situation(id)
            else:
                raise RuntimeError("Unexpected invalidation action")
        else:
            raise RuntimeError("Unexpected invalidation target")

        return gRPC.Empty()


class Context(object):

    def __init__(self, name, host="localhost",port=50051):

        self.name = name
        self.worlds = WorldsProxy(self)

        self.invalidation_port = 0

        while self.invalidation_port == 0:
            invalidation_port = random.randint(port + 1,60000)
            logger.info("Creating my own invalidation server on port %s..." % (invalidation_port))

            self.invalidation_server = gRPC.beta_create_UnderworldsInvalidation_server(InvalidationServer(self))
            self.invalidation_port = self.invalidation_server.add_insecure_port('[::]:%d' % invalidation_port)

            if self.invalidation_port == 0:
                logger.error("The port for my invalidation server is already in use! Trying another one...")


        self.invalidation_server.start()

        logger.info("Invalidation server created")


        if "UWDS_SERVER" in os.environ and os.environ["UWDS_SERVER"] != "":
            if ":" in os.environ["UWDS_SERVER"]:
                host, port = os.environ["UWDS_SERVER"].split(":")
                port = int(port)
            else:
                host = os.environ["UWDS_SERVER"]

        logger.info("Connecting to the underworlds server on %s:%s..." % (host, port))


        try:
            channel = implementations.insecure_channel(host, port)
            self.rpc = gRPC.beta_create_Underworlds_stub(channel)

            self.id = self.rpc.helo(gRPC.Welcome(name=name,
                                                 host="localhost", 
                                                 invalidation_server_port=self.invalidation_port), _TIMEOUT_SECONDS).id
        except NetworkError as e:
            logger.fatal("Underworld server unreachable on %s:%d! Is it started?\n"
                         "Set UWDS_SERVER=host:port if underworlded is running on a different machine.\n"
                         "Original error: %s" % (host, port, str(e)))
            sys.exit(1)
        except AbortionError as e:
            logger.fatal("Underworld server refused connection on %s:%d.\n"
                         "Original error: %s" % (host, port, str(e)))
            sys.exit(1)

        logger.info("<%s> connected to the underworlds server." % self.name)


    def reset(self):
        """ Hard reset of Underworlds: all the worlds are deleted.
        The existing mesh database is kept, however.
        This does not impact the list of known clients (ie, clients do not have to
        call 'helo' again).
        """
        self.rpc.reset(gRPC.Client(id=self.id), _TIMEOUT_SECONDS)


    def topology(self):
        """Returns the current topology to the underworlds environment.

        It returns an object with two members:
        - 'worlds': the list of all worlds' names known to the system
        - 'clients': the list of known clients. Each client is an object with
          the following members:
          - client.id
          - client.name
          - client.links: a list of the 'links' between this client and the
            worlds. Each link has the following members:
            - link.world
            - link.type: `READER`, `PROVIDER`, `FILTER`, `MONITOR`, see `types.py`
            - link.last_activity: the timestamp of the last time this link has been used
        """
        return self.rpc.topology(gRPC.Client(id=self.id), _TIMEOUT_SECONDS)

    def uptime(self):
        """Returns the server uptime in seconds.
        """
        return self.rpc.uptime(gRPC.Client(id=self.id),_TIMEOUT_SECONDS).time

    def has_mesh(self, id):
        ok = self.rpc.hasMesh(gRPC.MeshInContext(client=gRPC.Client(id=self.id),
                                                   mesh=gRPC.Mesh(id=id)),
                              _TIMEOUT_SECONDS)
        return ok.value

    def mesh(self, id):
        mesh = self.rpc.getMesh(gRPC.MeshInContext(client=gRPC.Client(id=self.id),
                                                   mesh=gRPC.Mesh(id=id)),
                              _TIMEOUT_SECONDS_MESH_LOADING)
        return Mesh.deserialize(mesh)

    def push_mesh(self, mesh):

        starttime = time.time()
        try:
            self.rpc.pushMesh(gRPC.MeshInContext(client=gRPC.Client(id=self.id),
                                             mesh=mesh.serialize(gRPC.Mesh)),
                              _TIMEOUT_SECONDS_MESH_LOADING)
        except ExpirationError:
            logger.error("Timeout while trying to push a mesh to the server!")

        logger.info("Pushed mesh <%s> in %.2fsec" % (mesh.id, time.time() - starttime))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        logger.info("Closing context [%s]..." % self.name)
        self.rpc.byebye(gRPC.Client(id=self.id), _TIMEOUT_SECONDS)
        self.invalidation_server.stop(1).wait()
        logger.info("The context [%s] is now closed." % self.name)

    def __repr__(self):
        return "Underworlds context for " + self.name



from functools import wraps
import time

PROFILING_ENABLED=False

def profile(f):

    if not PROFILING_ENABLED:
        return f

    try:
        name = f.__qualname__ # python3
    except:
        name = f.__module__ + "." + f.__name__ # python3

    @wraps(f)
    def wrapper(*args, **kwargs):

        stime=time.time();print("PROFILE;%f;enter %s" % (stime,name))

        res=f(*args, **kwargs)

        endtime=time.time();print("PROFILE;%f;exit %s;%.2f" % (endtime,name,(endtime-stime)*1000))
        return res

    return wrapper

def profileonce(msg):
    if not PROFILING_ENABLED:
        return
    print("PROFILE;%f;%s" % (time.time(),msg))



#!/usr/bin/env python
#-*- coding: UTF-8 -*-

import sys
import uuid

import logging; logger = logging.getLogger("underworlds.model_loader")

try:
    from pyassimp import core as pyassimp
    from pyassimp.postprocess import *
except BaseException as ae:
    logger.error("pyassimp could not be loaded: %s" % ae)
    pass

import math
import underworlds
from underworlds.types import *
from underworlds.helpers import transformations

DEFAULT_WORLD = "base"
ROTATION_180_X = numpy.array([[1,0,0,0],[0,-1,0,0],[0,0,-1,0],[0,0,0,1]], dtype=numpy.float32)

class ModelLoader:

    def __init__(self):

        self.meshes = {}

        # mapping {assimp name: (assimp node, underworld node)}
        self.node_map = {}

    def node_boundingbox(self, node):
        """ Returns the AABB bounding box of an ASSIMP node.
        Be careful: this is the *untransformed* bounding box,
        ie, the bounding box of the mesh in the node frame.
        """
        bb_min = [1e10, 1e10, 1e10] # x,y,z
        bb_max = [-1e10, -1e10, -1e10] # x,y,z
        for mesh in node.meshes:
            for v in mesh.vertices:
                bb_min[0] = round(min(bb_min[0], v[0]), 5)
                bb_min[1] = round(min(bb_min[1], v[1]), 5)
                bb_min[2] = round(min(bb_min[2], v[2]), 5)
                bb_max[0] = round(max(bb_max[0], v[0]), 5)
                bb_max[1] = round(max(bb_max[1], v[1]), 5)
                bb_max[2] = round(max(bb_max[2], v[2]), 5)
        return (bb_min, bb_max)

    def fill_node_details(self, 
                          assimp_node, 
                          underworlds_node, 
                          assimp_model,
                          custom_root=None):

        logger.debug("Parsing node " + str(assimp_node))
        underworlds_node.name = assimp_node.name

        if assimp_node is not assimp_model.rootnode:
            parent = assimp_node.parent
            if parent == assimp_model.rootnode and custom_root:
                underworlds_node.parent = custom_root.id
            else:
                underworlds_node.parent = self.node_map[assimp_node.parent.name][1].id

        # No need to specify the node's children: this is automatically done
        # by underworlds

        underworlds_node.transformation = assimp_node.transformation.astype(numpy.float32)

        if assimp_node.meshes:
            underworlds_node.type = MESH
            underworlds_node.cad = []
            underworlds_node.hires = []

            for m in assimp_node.meshes:
                
                mesh = Mesh(m.vertices.tolist(), 
                            m.faces.tolist(), 
                            m.normals.tolist(),
                            m.material.properties["diffuse"])

                id = mesh.id
                logger.debug("\tLoading mesh %s" % id)
                self.meshes[id] = mesh
                underworlds_node.cad.append(id)
                underworlds_node.hires.append(id)

            underworlds_node.aabb = self.node_boundingbox(assimp_node)

        elif assimp_node.name in [c.name for c in assimp_model.cameras]:
            logger.debug("\tAdding camera <%s>" % assimp_node.name)

            [cam] = [c for c in assimp_model.cameras if c.name == assimp_node.name]
            underworlds_node.type = CAMERA
            underworlds_node.clipplanenear = cam.clipplanenear
            underworlds_node.clipplanefar = cam.clipplanefar

            if numpy.allclose(cam.lookat, [0,0,-1]) and numpy.allclose(cam.up, [0,1,0]): # Cameras in .blend files

                # Rotate by 180deg around X to have Z pointing forward
                underworlds_node.transformation = numpy.dot(underworlds_node.transformation, ROTATION_180_X)
            else:
                raise RuntimeError("I do not know how to normalize this camera orientation: lookat=%s, up=%s" % (cam.lookat, cam.up))

            if cam.aspect == 0.0:
                logger.warning("Camera aspect not set. Setting to default 4:3")
                underworlds_node.aspect = 1.333
            else:
                underworlds_node.aspect = cam.aspect

            underworlds_node.horizontalfov = cam.horizontalfov

            #underworlds_node.lookat = [round(a, 5) for a in cam.lookat]
        else:
            underworlds_node.type = ENTITY

    def recur_node(self, assimp_node, model, level = 0):
        logger.info("  " + "\t" * level + "- " + str(assimp_node))

        if assimp_node != model.rootnode: # the rootnode is already there
            self.node_map[assimp_node.name] = (assimp_node, Node()) # cannot use assimp_node as key: it is unhashable
        for child in assimp_node.children:
            self.recur_node(child, model, level + 1)


    def load(self, filename, ctx=None, world=DEFAULT_WORLD, root=None, only_meshes=False):
        """Loads a Collada (or any Assimp compatible model) file in the world.

        The kinematic chains are added to the world's geometric state.
        The meshes are added to the meshes repository.

        A new 'load' event is also added in the world timeline.

        :param string path: the path (relative or absolute) to the Collada resource
        :param Context ctx: an existing underworlds context. If not provided, a
                            new one is created (named 'model loader')
        :param string world: the target world for the creation of nodes
        :param Node root: if given, the loaded nodes will be parented to this
                          node instead of the scene's root.
        :param bool only_meshes: if true, no node is created. Only the
        meshes are pushed to the server.
        :returns: the list of loaded underworlds nodes.
        """

        self.meshes = {}
        self.node_map = {}

        if not only_meshes and world is None:
            raise RuntimeError("Can not create nodes if the world is None")

        close_ctx_at_end = False

        # Create a context if needed:
        if ctx is None:
            close_ctx_at_end = True
            ctx = underworlds.Context("model loader")

        logger.info("Loading model <%s> with ASSIMP..." % filename)
        model = pyassimp.load(filename, aiProcessPreset_TargetRealtime_MaxQuality)
        logger.info("...done")

        if not only_meshes:
            if not root:
                logger.info("Merging the root nodes")
                self.node_map[model.rootnode.name] = (model.rootnode, ctx.worlds[world].scene.rootnode)
        else:
            self.node_map[model.rootnode.name] = (model.rootnode, Node())

        logger.info("Nodes found:")
        self.recur_node(model.rootnode, model)

        logger.info("%d nodes in the model" % len(self.node_map))
        logger.info("Loading the nodes...")
        for n, pair in list(self.node_map.items()):
            self.fill_node_details(*pair,
                                    assimp_model = model,
                                    custom_root=root)
        logger.info("...done")


        # Send first the meshes to make sure they are available on the server
        # when needed by clients.  (but only if they do not already exist on
        # the server)
        logger.info("Sending meshes to the server...")
        count_sent = 0
        count_notsent = 0
        for id, mesh in list(self.meshes.items()):
            if ctx.has_mesh(id):
                logger.debug("Not resending mesh <%s>: already available on the server" % id)
                count_notsent += 1
            else:
                logger.debug("Sending mesh <%s>" % id)
                ctx.push_mesh(mesh)
                count_sent += 1

        logger.info("Sent %d meshes (%d were already available on the server)" % (count_sent, count_notsent))

        if not only_meshes:
            nodes = ctx.worlds[world].scene.nodes

            # Send the nodes to the server (only the nodes)
            logger.info("Sending the nodes to the server...")
            for name, pair in list(self.node_map.items()):
                nodes.update(pair[1])


        pyassimp.release(model)



        if close_ctx_at_end:
            ctx.close()

        return [nodes[1] for _,nodes in self.node_map.items()]

    def load_meshes(self, filename, ctx=None):
        """Pushes meshes from any Assimp-compatible 3D model to the server's
        mesh repository.

        A new 'load' event is also added in the world timeline.

        :param string path: the path (relative or absolute) to the 3D model
        :param Context ctx: an existing underworlds context. If not provided, a
                            new one is created (named 'model loader')
        :returns: a dictionary {mesh name: mesh ID} 

        :see: `load` loads the meshes and creates corresponding nodes.
        """

        return {n.name:n.cad for n in self.load(filename, ctx, None, None, True) if n.type == MESH}

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    if len(sys.argv) == 1:
        print(("usage: %s path/to/model" % sys.argv[0]))
        sys.exit(2)

    # Loads the model in the 'DEFAULT_WORLD' (ie, 'base')
    ModelLoader().load(sys.argv[1])
